#!/usr/bin/perl

# Author : Damien KROTKINE (damien@tuxfamily.org)
#
# Contributors :
#
# Copyright (C) 2002 Damien KROTKINE (damien@tuxfamily.org)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; only version 2, not any other version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

=head1 NAME

PERL : Libconf::Glueconf::Samba::Smb_conf
C/PYTHON/RUBY : Samba/Smb_conf

=head1 DESCRIPTION

Libconf::Glueconf::Samba::Smb_conf is a template that handles the samba smb.conf configuration file.

=head1 SYNOPSIS

  $samba = new Libconf::Glueconf::Samba::Smb_conf({ filename => '/etc/samba/smb.conf' });
  $samba->{homes}->{writable} = 'TEST';
  $samba->{share}->{test} = 'yes';
  $samba->write_conf('/etc/samba/edited_smb.conf');

=head1 CONSTRUCTOR

=over

=item B<new($options)>

creates the template

  $samba = new Libconf::Glueconf::Samba::Smb_conf({
                                                   .. arguments .. 
                                                 })

the options are the same as in L<Libconf::Templates::Samba>, look
there to have the details.

=back

=head1 METHODS

See <Libconf::Glueconf> for general methods

=cut

package Libconf::Glueconf::Samba::Smb_conf;
use strict;
use vars qw(@ISA);
use Libconf;
use Libconf::Glueconf;
use Libconf::Templates::Samba;
@ISA = qw(Libconf::Glueconf);

sub new {
    my ($class, $args) = @_;
    my $libconf = new Libconf::Templates::Samba(init_args($args, { glueconf_template_name => 'Samba/Smb_conf' } ));
    $libconf->read_conf();
    $libconf->set_uniq(qw(key sections));

    tie my %wrapper, 'Libconf::Glueconf::Samba::Smb_conf::Wrapper', $libconf;
    bless \%wrapper, $class;
}

sub description { 'template for samba smb.conf configuration file' }

package Libconf::Glueconf::Samba::Smb_conf::Wrapper;
use Libconf::Glueconf::Wrapper::Hash;
our @ISA = qw(Libconf::Glueconf::Wrapper::Hash);

sub debug { Libconf::debug(@_) }

sub DELETE {
    my ($obj, $key) = @_;
    debug("key: $key");
    my @pos = $obj->{libconf}->find_atom_pos( { type => 'SECTION', sections => [], section_name => $key });
    foreach (@pos) {
        $obj->{libconf}->delete_section($_, 1);
    }
}

sub FIRSTKEY {
    my ($obj) = @_;
    debug();
    my $atom = $obj->{libconf}->get_atom(0);
    $atom->{type} eq 'SECTION' or die "atom n 0 is not a section\n";
    $atom->{section_name};
}

sub EXISTS {
    my ($obj, $key) = @_;
    debug("key : $key");
    my $pos = $obj->{libconf}->find_atom_pos( { type => 'SECTION', sections => [], section_name => $key });
    defined $pos;
}

sub NEXTKEY {
    my ($obj, $lastkey) = @_;
    debug("lastkey : $lastkey");
    my @pos = $obj->{libconf}->find_atom_pos( { type => 'SECTION', sections => [] });
    my $i = 0;
    my $index;
    foreach (@pos) {
        $obj->{libconf}->get_atom($_)->{section_name} eq $lastkey and $index = $i;
        $i++;
    }
    defined $index or return undef;
    $index == $#pos and return undef;
    $obj->{libconf}->get_atom($pos[$index+1])->{section_name};
}

sub STORE {
    my ($obj, $key, $value) = @_;
    debug("key : $key - value : $value");
    ref $value eq 'HASH' or die 'trying to store something else than hash ref';
    my $index;
    my $pos = $obj->{libconf}->find_atom_pos({ type => 'SECTION', section_name => $key });
    defined $pos or $pos = $obj->{libconf}->append_atom({ section_name => $key, type => 'SECTION' });
    my %hash;
    tie %hash, 'Libconf::Glueconf::Samba::Smb_conf::SectionWrapper', $obj->{libconf}, $pos;
    %hash = %$value;
}

sub FETCH {
    my ($obj, $key) = @_;
    debug("key : $key");
    $key eq 'libconf' and return $obj->{libconf};
    substr($key, 0, 1) eq '_' and return $obj->{$key};
    my $section_position = $obj->{libconf}->find_atom_pos( { type => 'SECTION', section_name => $key, sections => [ ] });
    $obj->{libconf}{_fetch_atom_pos} and $obj->{libconf}{_atom_pos} = $section_position;
    defined $section_position or $section_position = $obj->{libconf}->append_atom({ section_name => $key, type => 'SECTION' });
    my %ret;
    tie %ret, 'Libconf::Glueconf::Samba::Smb_conf::SectionWrapper', $obj->{libconf}, $section_position;
    \%ret;
}

package Libconf::Glueconf::Samba::Smb_conf::SectionWrapper;
use Libconf::Glueconf::Wrapper::Hash;
our @ISA = qw(Libconf::Glueconf::Wrapper::Hash);

sub debug { Libconf::debug(@_) }

sub TIEHASH {
    my ($pkg, $libconf, $beginning) = @_;
    debug("beginning : $beginning");
    my $atom = $libconf->get_atom($beginning);
    my @depth = @{$atom->{sections}};
    scalar(keys %{$depth[0]}) == 0 and shift @depth;
    push @depth, { name => $atom->{section_name} };

    my $key = $beginning+1;
    foreach ($beginning+1..$libconf->get_size()-1) {
        my $atom = $libconf->get_atom($_);
        @{$atom->{sections}} > @depth and next;
        Libconf::compare_depth($atom->{sections}, \@depth) or $key--, last;
    } continue {
        $key++;
    }
    $key >= $libconf->get_size() and $key--;
    my %hash = (
                libconf => $libconf,
                firstatom => $beginning,
                sections => \@depth,
                lastatom => $key,
               );
    debug("firstatom : $hash{firstatom} - lastatom : $hash{lastatom}");
    return bless \%hash, $pkg;
}

sub CLEAR {
    my ($obj) = @_;
    debug();
    $obj->{lastatom} -= $obj->{libconf}->clear_section($obj->{firstatom}, 1);
    $obj->{lastatom} == $obj->{firstatom} or die "removed atoms number is wrong";
}

sub DELETE {
    my ($obj, $key) = @_;
    debug("key : $key");
    my @pos = $obj->{libconf}->find_atom_pos( {type => 'KEY_VALUE', key => $key },
                                              $obj->{firstatom}+1, $obj->{lastatom}
                                            );
    foreach (@pos) {
        $obj->{libconf}->delete_atom($_);
        $obj->{lastatom}--;
    }
}

sub FIRSTKEY {
    my ($obj) = @_;
    debug();
    $obj->{firstatom} == $obj->{lastatom} and return undef;
    my $atom = $obj->{libconf}->get_atom($obj->{firstatom}+1);
    $atom->{type} eq 'KEY_VALUE' and return $atom->{key};
    die "the first atom of the section is not a KEY_VALUE, I don't know how this is possible . section [" . $obj->{firstatom}->{section_name} ."]\n";
}

sub EXISTS {
    my ($obj, $key) = @_;
    debug("key : $key");
    my $pos = $obj->{libconf}->find_atom_pos( {type => 'KEY_VALUE', key => $key },
                                            $obj->{firstatom}+1, $obj->{lastatom}
                                          );
    defined $pos ? 1 : 0;
}

sub NEXTKEY {
    my ($obj, $lastkey) = @_;
    debug("SectionWrapper - NEXTKEY - lastkey : $lastkey");
    my $pos = $obj->{libconf}->find_atom_pos( {type => 'KEY_VALUE', key => $lastkey },
                                            $obj->{firstatom}+1, $obj->{lastatom}-1
                                          );
    defined $pos or return undef;
    $obj->{libconf}->get_atom($pos+1)->{key};
}

sub STORE {
    my ($obj, $key, $value) = @_;
    debug("key : $key - value : $value");
    #ref $value eq 'HASH' or die 'trying to store anything else than hash ref';
    $obj->check_type($value, 'SCALAR');
    my $pos = $obj->{libconf}->find_atom_pos( {type => 'KEY_VALUE', key => $key },
                                            $obj->{firstatom}+1, $obj->{lastatom}
                                          );
    if (!defined $pos) {
        $obj->{lastatom}++;
        $obj->{libconf}->insert_atom($obj->{lastatom}, { type => 'KEY_VALUE', key => $key,
                                                         sections => [ @{$obj->{sections}} ],
                                                       }
                                    );
        $pos = $obj->{lastatom};
    }
    $obj->store($pos, $value);
}

sub FETCH {
    my ($obj, $key) = @_;
    debug("key : $key");
    my $pos = $obj->{libconf}->find_atom_pos( {type => 'KEY_VALUE', key => $key },
                                            $obj->{firstatom}+1, $obj->{lastatom}
                                          );
    defined $pos or return undef;
#    $obj->fetch($obj->{libconf}->get_atom($pos), 'value');
    $obj->fetch($pos);
}

1;

